"""
Apps Visualizer apps state
"""
from typing import Optional
import asyncio
from datetime import datetime, timezone

from hopeit.app.context import EventContext
from hopeit.server.logger import engine_extra_logger

from hopeit.config_manager import RuntimeApps
from hopeit.config_manager.client import get_apps_config

from hopeit.apps_visualizer import AppsVisualizerSettings

logger, extra = engine_extra_logger()

_lock = asyncio.Lock()
_apps: Optional[RuntimeApps] = None
_expire: float = 0.0


async def get_runtime_apps(context: EventContext,
                           *, refresh: bool = False,
                           expand_events: bool = False) -> RuntimeApps:
    """
    Extract current runtime app_config objects
    """
    global _apps, _expire
    if not refresh and _lock.locked():
        raise RuntimeError("Events graph request in process. Ignoring")
    settings = context.settings(key="apps_visualizer", datatype=AppsVisualizerSettings)
    now_ts = datetime.now(tz=timezone.utc).timestamp()
    async with _lock:
        if _apps is None or refresh or now_ts > _expire:
            logger.info(context, "Contacting hosts config-manager...")
            _apps = await get_apps_config(settings.hosts, context, expand_events=expand_events)
            _expire = now_ts + settings.refresh_hosts_seconds
        else:
            logger.info(context, "Using cached runtime apps information.")
        return _apps



"""
Apps Visualizer plugin module
"""
from hopeit.dataobjects import dataclass, dataobject


@dataobject
@dataclass
class AppsVisualizerSettings:
    """
    Apps visualizer settings.

    Helper dataclasses to load "apps-visualizer" env section from plugin config.

    :field: hosts, str: comma-separated list of `http://host:port` entries to contact to query
        config-manager plugin for running apps layout
    :field: refresh_hosts_seconds: int, default 60: interval to contact nodes config-manager
    :field: live_recent_treshold_seconds, int, default 10: number of seconds to consider a node
        recently activated (the complete box will be highlighted)
    :field: live_active_treshold_seconds, int, default: 60: number of seconds to consider a node
        has been activated (border will be highlighted)
    """
    hosts: str
    refresh_hosts_seconds: int = 60
    live_recent_treshold_seconds: int = 10
    live_active_treshold_seconds: int = 60



"""
Events graph showing events, stream and dependencies for specified apps
"""
from typing import Optional

from hopeit.app.context import EventContext
from hopeit.app.api import event_api
from hopeit.dataobjects import dataclass, dataobject
from hopeit.app.events import collector_step, Collector
from hopeit.app.logger import app_extra_logger

from hopeit.config_manager import RuntimeApps, RuntimeAppInfo

from hopeit.apps_visualizer.apps import get_runtime_apps
from hopeit.apps_visualizer.graphs import Edge, Node, Graph, add_app_connections, get_edges, get_nodes
from hopeit.apps_visualizer.site.visualization import CytoscapeGraph, VisualizationOptions, \
    visualization_options, visualization_options_api_args  # noqa: F401  # pylint: disable=unused-import

logger, extra = app_extra_logger()

__steps__ = [
    'visualization_options',
    collector_step(payload=VisualizationOptions).gather(
        'runtime_apps',
        'config_graph',
        'cytoscape_data'
    ),
    'build_visualization'
]


@dataobject
@dataclass
class EventsGraphResult:
    runtime_apps: RuntimeApps
    graph: CytoscapeGraph
    options: VisualizationOptions


__api__ = event_api(
    summary="App Visualizer: Events Graph Data",
    description="App Visualizer: Events Graph Data",
    query_args=visualization_options_api_args(),
    responses={
        200: (EventsGraphResult, "Graph Data with applied Live Stats")
    }
)


async def runtime_apps(collector: Collector, context: EventContext) -> RuntimeApps:
    """
    Extract current runtime app_config objects
    """
    options: VisualizationOptions = await collector['payload']
    return await get_runtime_apps(context, expand_events=options.expanded_view)


def _filter_apps(runtime_info: RuntimeAppInfo, options: VisualizationOptions) -> bool:
    return options.app_prefix == runtime_info.app_config.app.name[0:len(options.app_prefix)]


def _filter_hosts(runtime_info: RuntimeAppInfo, options: VisualizationOptions) -> bool:
    return (
        options.host_filter == '' or any(
            options.host_filter in server.url for server in runtime_info.servers
        )
    )


async def config_graph(collector: Collector, context: EventContext) -> Optional[Graph]:
    """
    Generates Graph object with nodes and edges from server runtime active configuration
    """
    options: VisualizationOptions = await collector['payload']
    all_apps: RuntimeApps = await collector['runtime_apps']

    filtered_apps = [
        (app_key, runtime_info)
        for app_key, runtime_info in all_apps.apps.items()
        if _filter_apps(runtime_info, options) and _filter_hosts(runtime_info, options)
    ]

    events = {
        event_name: event_info
        for _, app_info in filtered_apps
        for event_name, event_info in app_info.effective_events.items()
    }

    app_connections = {
        f"{app_key}.{app_conn_key}": app_connection
        for app_key, app_info in filtered_apps
        for app_conn_key, app_connection in app_info.app_config.app_connections.items()
    }

    nodes = get_nodes(events, expanded_view=options.expanded_view)
    add_app_connections(nodes, app_connections=app_connections, events=events)
    edges = get_edges(nodes)
    return Graph(nodes=list(nodes.values()), edges=edges)


async def cytoscape_data(collector: Collector, context: EventContext) -> CytoscapeGraph:
    """
    Converts Graph to cytoscape json format
    """
    def _edge_label(edge: Edge) -> str:
        label = edge.label.split('.')[-1]
        if label in ("AUTO", "POST", "GET", "MULTIPART"):
            return ""
        return label

    def _node_label(node: Node) -> str:
        comps = node.label.split(".")
        if len(comps) > 2:
            return '\n'.join(['.'.join(comps[0:2]), '.'.join(comps[2:])])
        return node.label

    graph: Graph = await collector['config_graph']

    nodes = {
        node.id: {"data": {
            "id": node.id,
            "content": _node_label(node),
        }, "classes": node.type.value}
        for node in graph.nodes
    }
    edges = {
        f"edge_{edge.id}": {"data": {
            "id": f"edge_{edge.id}",
            "source": edge.source,
            "target": edge.target,
            "label": _edge_label(edge)
        }}
        for edge in graph.edges
    }
    return CytoscapeGraph(data={**nodes, **edges})


async def build_visualization(collector: Collector, context: EventContext) -> EventsGraphResult:
    return EventsGraphResult(
        runtime_apps=await collector['runtime_apps'],
        graph=await collector['cytoscape_data'],
        options=await collector['payload']
    )



"""
Collect Event Stats from log stream
"""
from typing import Deque, Dict, Set, Tuple
from collections import defaultdict, deque
from datetime import datetime, timedelta, timezone

from hopeit.app.context import EventContext
from hopeit.dataobjects import dataclass, dataobject
from hopeit.server.names import auto_path

from hopeit.log_streamer import LogBatch, LogEntry


@dataobject
@dataclass
class EventStats:
    started: int = 0
    done: int = 0
    failed: int = 0
    ignored: int = 0
    recent: int = 0

    @property
    def pending(self):
        return max(0, self.started - self.done - self.failed - self.ignored)


recent_entries: Deque[LogEntry] = deque(maxlen=2000)

__steps__ = ['consume']


async def consume(batch: LogBatch, context: EventContext) -> None:
    recent_entries.extend(batch.entries)


def get_stats(host_pids: Set[Tuple[str, str]], time_window_secs: int, recent_secs: int) -> Dict[str, EventStats]:
    """
    Collect node event stats, counting for each node with recent entries
    number of started, done, pending and failed events.

    :param: host_pids, str tuple, set of host-pid pairs to filter log entries
    :param: time_window_secs: int, consider events not older than this
    :param: recent_secs: int, events on the last seconds will be counted as recent
    """
    stats: Dict[str, EventStats] = defaultdict(EventStats)
    now_ts = datetime.now(tz=timezone.utc)
    from_ts = (now_ts - timedelta(seconds=time_window_secs)).strftime("%Y-%m-%d %H:%M:%S")
    recent_ts = (now_ts - timedelta(seconds=recent_secs)).strftime("%Y-%m-%d %H:%M:%S")
    for entry in recent_entries:
        if entry.ts >= from_ts and (entry.host, entry.pid) in host_pids:
            event_name = entry.extra.get("event.app", auto_path(entry.app_name, entry.app_version)) + '.'
            plugin = entry.extra.get("event.plugin")
            if plugin:
                event_name += plugin + '.'
            event_name += entry.extra.get("stream.event_name", entry.event_name)
            event_stats = stats[event_name]
            stream_key = ">" + entry.extra.get("stream.name", "NA")
            stream_queue = "." + entry.extra.get("stream.queue", "")
            if stream_key[-len(stream_queue):] == stream_queue:
                stream_key = stream_key[0:-len(stream_queue)]
            stream_stats = stats[stream_key]
            stream_key = (stream_key + stream_queue).strip('.')
            stats[stream_key] = stream_stats
            if entry.msg == "START":
                event_stats.started += 1
                stream_stats.started += 1
            elif entry.msg == "DONE":
                event_stats.done += 1
                stream_stats.done += 1
            elif entry.msg == "FAILED":
                event_stats.failed += 1
                stream_stats.failed += 1
            elif entry.msg == "IGNORED":
                event_stats.ignored += 1
                stream_stats.ignored += 1
            if entry.ts >= recent_ts:
                event_stats.recent += 1
                stream_stats.recent += 1
    return stats



"""
Visualization metadata
"""
from hopeit.app.context import EventContext
from typing import Dict, Any, Optional
from dataclasses import dataclass, field

from hopeit.dataobjects import dataobject


@dataobject
@dataclass
class CytoscapeGraph:
    data: Dict[str, Dict[str, Any]] = field(default_factory=dict)


@dataobject
@dataclass
class VisualizationOptions:
    app_prefix: str = ''
    host_filter: str = ''
    expanded_view: bool = False
    live: bool = False


def visualization_options_api_args():
    return [
        ("app_prefix", Optional[str], "app name prefix to filter"),
        ("host_filter", Optional[str], "host name filter substring"),
        ("expanded_view", Optional[bool], "if `true` shows each stream queue as a separated stream"),
        ("live", Optional[bool], "if `true` enable live stats refreshing")
    ]


async def visualization_options(payload: None, context: EventContext,
                                *, app_prefix: str = '',
                                host_filter: str = '',
                                expanded_view: bool = False,
                                live: bool = False) -> VisualizationOptions:
    return VisualizationOptions(
        app_prefix=app_prefix,
        host_filter=host_filter,
        expanded_view=expanded_view is True or expanded_view == 'true',
        live=live is True or live == 'true'
    )
